@@ -46,12 +46,12 @@
 
 void decide_column(size_t ncols_numeric, size_t ncols_categ, size_t &col_chosen, ColType &col_type,
                    RNG_engine &rnd_generator, std::uniform_int_distribution<size_t> &runif,
-                   std::discrete_distribution<size_t> &col_sampler)
+                   WeightedColSampler &col_sampler)
 {
-    if (!col_sampler.max())
+    if (!col_sampler.is_initialized())
         col_chosen = runif(rnd_generator);
     else
-        col_chosen = col_sampler(rnd_generator);
+        col_chosen = col_sampler.sample_col(rnd_generator);
 
     if (col_chosen >= ncols_numeric)
     {
@@ -64,21 +64,29 @@ void decide_column(size_t ncols_numeric, size_t ncols_categ, size_t &col_chosen,
 
 void add_unsplittable_col(WorkerMemory &workspace, IsoTree &tree, InputData &input_data)
 {
-    if (tree.col_type == Numeric)
+    if (tree.col_type == Numeric) {
         workspace.cols_possible[tree.col_num] = false;
-    else
+        workspace.col_sampler.drop_col(tree.col_num);
+    }
+    else {
         workspace.cols_possible[tree.col_num + input_data.ncols_numeric] = false;
+        workspace.col_sampler.drop_col(tree.col_num + input_data.ncols_numeric);
+    }
 }
 
 void add_unsplittable_col(WorkerMemory &workspace, InputData &input_data)
 {
-    if (workspace.col_type == Numeric)
+    if (workspace.col_type == Numeric) {
         workspace.cols_possible[workspace.col_chosen] = false;
-    else
+        workspace.col_sampler.drop_col(workspace.col_chosen);
+    }
+    else {
         workspace.cols_possible[workspace.col_chosen + input_data.ncols_numeric] = false;
+        workspace.col_sampler.drop_col(workspace.col_chosen + input_data.ncols_numeric);
+    }
 }
 
-bool check_is_not_unsplittable_col(WorkerMemory &workspace, IsoTree &tree, InputData &input_data)
+bool check_is_splittable_col(WorkerMemory &workspace, IsoTree &tree, InputData &input_data)
 {
     if (tree.col_type == Numeric)
         return workspace.cols_possible[tree.col_num];
@@ -152,21 +160,6 @@ int choose_cat_from_present(WorkerMemory &workspace, InputData &input_data, size
     return -1; /* this will never be reached, but CRAN complains otherwise */
 }
 
-void update_col_sampler(WorkerMemory &workspace, InputData &input_data)
-{
-    if (!workspace.col_sampler.max())
-        return;
-
-    std::vector<double> col_weights = workspace.col_sampler.probabilities();
-    for (size_t col = 0; col < input_data.ncols_numeric; col++)
-        if (!workspace.cols_possible[col])
-            col_weights[col] = 0;
-    for (size_t col = 0; col < input_data.ncols_categ; col++)
-        if (!workspace.cols_possible[col + input_data.ncols_numeric])
-            col_weights[col + input_data.ncols_numeric] = 0;
-    workspace.col_sampler = std::discrete_distribution<size_t>(col_weights.begin(), col_weights.end());
-}
-
 bool is_col_taken(std::vector<bool> &col_is_taken, std::unordered_set<size_t> &col_is_taken_s,
                   InputData &input_data, size_t col_num, ColType col_type)
 {
@@ -268,57 +261,68 @@ void remap_terminal_trees(IsoForest *model_outputs, ExtIsoForest *model_outputs_
     }
 }
 
-void backup_recursion_state(WorkerMemory &workspace, RecursionState &recursion_state)
+RecursionState::RecursionState(WorkerMemory &workspace, bool full_state)
 {
-    recursion_state.st            = workspace.st;
-    recursion_state.st_NA         = workspace.st_NA;
-    recursion_state.end_NA        = workspace.end_NA;
-    recursion_state.split_ix      = workspace.split_ix;
-    recursion_state.end           = workspace.end;
-    recursion_state.cols_possible = workspace.cols_possible;
-    recursion_state.col_sampler   = workspace.col_sampler;
+    this->full_state = full_state;
 
-    /* for the extended model, it's not necessary to copy everything */
-    if (!workspace.comb_val.size())
+    this->split_ix         =  workspace.split_ix;
+    this->end              =  workspace.end;
+    this->cols_possible    =  workspace.cols_possible;
+    this->go_to_shuffle    =  workspace.go_to_shuffle;
+    if (workspace.col_sampler.is_initialized())
+        this->col_sampler  =  workspace.col_sampler;
+
+    if (this->full_state)
     {
-        recursion_state.ix_arr = std::vector<size_t>(workspace.ix_arr.begin() + workspace.st_NA,
-                                                     workspace.ix_arr.begin() + workspace.end + 1);
-        size_t tot = workspace.end - workspace.st_NA + 1;
-        if (workspace.weights_arr.size() || workspace.weights_map.size())
-            recursion_state.weights_arr = std::unique_ptr<double[]>(new double[tot]);
-        if (workspace.weights_arr.size())
-            for (size_t ix = 0; ix < tot; ix++)
-                recursion_state.weights_arr[ix] = workspace.weights_arr[workspace.ix_arr[ix + workspace.st_NA]];
-        else if (workspace.weights_map.size())
-            for (size_t ix = 0; ix < tot; ix++)
-                recursion_state.weights_arr[ix] = workspace.weights_map[workspace.ix_arr[ix + workspace.st_NA]];
-        
+        this->st            = workspace.st;
+        this->st_NA         = workspace.st_NA;
+        this->end_NA        = workspace.end_NA;
 
+        /* for the extended model, it's not necessary to copy everything */
+        if (!workspace.comb_val.size())
+        {
+            this->ix_arr = std::vector<size_t>(workspace.ix_arr.begin() + workspace.st_NA,
+                                                         workspace.ix_arr.begin() + workspace.end + 1);
+            size_t tot = workspace.end - workspace.st_NA + 1;
+            if (workspace.weights_arr.size() || workspace.weights_map.size())
+                this->weights_arr = std::unique_ptr<double[]>(new double[tot]);
+            if (workspace.weights_arr.size())
+                for (size_t ix = 0; ix < tot; ix++)
+                    this->weights_arr[ix] = workspace.weights_arr[workspace.ix_arr[ix + workspace.st_NA]];
+            else if (workspace.weights_map.size())
+                for (size_t ix = 0; ix < tot; ix++)
+                    this->weights_arr[ix] = workspace.weights_map[workspace.ix_arr[ix + workspace.st_NA]];
+        }
     }
 }
 
 
-void restore_recursion_state(WorkerMemory &workspace, RecursionState &recursion_state)
+void RecursionState::restore_state(WorkerMemory &workspace)
 {
-    workspace.st            = recursion_state.st;
-    workspace.st_NA         = recursion_state.st_NA;
-    workspace.end_NA        = recursion_state.end_NA;
-    workspace.split_ix      = recursion_state.split_ix;
-    workspace.end           = recursion_state.end;
-    workspace.cols_possible = std::move(recursion_state.cols_possible);
-    workspace.col_sampler   = std::move(recursion_state.col_sampler);
+    workspace.split_ix       =  this->split_ix;
+    workspace.end            =  this->end;
+    workspace.cols_possible  =  std::move(this->cols_possible);
+    workspace.col_sampler    =  std::move(this->col_sampler);
+    workspace.go_to_shuffle  =  this->go_to_shuffle;
 
-    if (!workspace.comb_val.size())
+    if (this->full_state)
     {
-        std::copy(recursion_state.ix_arr.begin(),
-                  recursion_state.ix_arr.end(),
-                  workspace.ix_arr.begin() + recursion_state.st_NA);
-        size_t tot = workspace.end - workspace.st_NA + 1;
-        if (workspace.weights_arr.size())
-            for (size_t ix = 0; ix < tot; ix++)
-                workspace.weights_arr[workspace.ix_arr[ix + workspace.st_NA]] = recursion_state.weights_arr[ix];
-        else if (workspace.weights_map.size())
-            for (size_t ix = 0; ix < tot; ix++)
-                workspace.weights_map[workspace.ix_arr[ix + workspace.st_NA]] = recursion_state.weights_arr[ix];
+        workspace.st            = this->st;
+        workspace.st_NA         = this->st_NA;
+        workspace.end_NA        = this->end_NA;
+
+        if (!workspace.comb_val.size())
+        {
+            std::copy(this->ix_arr.begin(),
+                      this->ix_arr.end(),
+                      workspace.ix_arr.begin() + this->st_NA);
+            size_t tot = workspace.end - workspace.st_NA + 1;
+            if (workspace.weights_arr.size())
+                for (size_t ix = 0; ix < tot; ix++)
+                    workspace.weights_arr[workspace.ix_arr[ix + workspace.st_NA]] = this->weights_arr[ix];
+            else if (workspace.weights_map.size())
+                for (size_t ix = 0; ix < tot; ix++)
+                    workspace.weights_map[workspace.ix_arr[ix + workspace.st_NA]] = this->weights_arr[ix];
+        }
     }
 }
