@@ -252,6 +252,9 @@ void split_itree_recursive(std::vector<IsoTree>     &trees,
         else
             workspace.criterion = NoCrit;
 
+        if (workspace.go_to_shuffle)
+            goto probe_all;
+
 
         /* pick column at random */
         decide_column(input_data.ncols_numeric, input_data.ncols_categ,
@@ -276,7 +279,7 @@ void split_itree_recursive(std::vector<IsoTree>     &trees,
                               trees.back().col_num, trees.back().col_type,
                               workspace.rnd_generator, workspace.runif,
                               workspace.col_sampler);
-                if (!check_is_not_unsplittable_col(workspace, trees.back(), input_data))
+                if (check_is_splittable_col(workspace, trees.back(), input_data))
                 {
                     get_split_range(workspace, input_data, model_params, trees.back());
                     if (!workspace.unsplittable)
@@ -288,103 +291,77 @@ void split_itree_recursive(std::vector<IsoTree>     &trees,
             }
             while (workspace.ncols_tried < input_data.ncols_tot / 2);
 
-            /* if that didn't work, then check all the columns that are still splittable */
+            /* if that didn't work, try to check all columns in random order */
             if (workspace.unsplittable)
             {
-                workspace.ncols_tried = 0; /* note: this is used here as a counter for the number of still splittable columns */
-                if (input_data.Xc_indptr == NULL)
+                probe_all:
+                workspace.go_to_shuffle = true;
+                if (!workspace.col_sampler.is_initialized())
                 {
-                    for (size_t col = 0; col < input_data.ncols_numeric; col++)
-                    {
-                        if (!workspace.cols_possible[col]) continue;
-                        get_range(workspace.ix_arr.data(), input_data.numeric_data + input_data.nrows * col,
-                                  workspace.st, workspace.end, model_params.missing_action,
-                                  workspace.xmin, workspace.xmax, workspace.unsplittable);
-                        workspace.cols_possible[col] = !workspace.unsplittable;
-                        workspace.ncols_tried += !workspace.unsplittable;
-                    }
+                    if (workspace.cols_shuffled.size() < input_data.ncols_tot)
+                        workspace.cols_shuffled.resize(input_data.ncols_tot);
+                    std::iota(workspace.cols_shuffled.begin(), workspace.cols_shuffled.end(), (size_t)0);
+                    std::shuffle(workspace.cols_shuffled.begin(), workspace.cols_shuffled.end(), workspace.rnd_generator);
                 }
 
                 else
                 {
-                    for (size_t col = 0; col < input_data.ncols_numeric; col++)
+                    workspace.col_sampler.shuffle_cols(workspace.cols_shuffled, workspace.rnd_generator);
+                }
+                
+                for (size_t col : workspace.cols_shuffled)
+                {
+                    if (!workspace.cols_possible[col]) continue;
+                    
+                    if (col < input_data.ncols_numeric)
                     {
-                        if (!workspace.cols_possible[col]) continue;
-                        get_range(workspace.ix_arr.data(), workspace.st, workspace.end, col,
-                                  input_data.Xc, input_data.Xc_ind, input_data.Xc_indptr,
-                                  model_params.missing_action, workspace.xmin, workspace.xmax, workspace.unsplittable);
-                        workspace.cols_possible[col] = !workspace.unsplittable;
-                        workspace.ncols_tried += !workspace.unsplittable;
+                        if (input_data.Xc_indptr == NULL)
+                            get_range(workspace.ix_arr.data(), input_data.numeric_data + input_data.nrows * col,
+                                      workspace.st, workspace.end, model_params.missing_action,
+                                      workspace.xmin, workspace.xmax, workspace.unsplittable);
+                        else
+                            get_range(workspace.ix_arr.data(), workspace.st, workspace.end, col,
+                                      input_data.Xc, input_data.Xc_ind, input_data.Xc_indptr,
+                                      model_params.missing_action, workspace.xmin, workspace.xmax, workspace.unsplittable);
                     }
-                }
 
-                for (size_t col = 0; col < input_data.ncols_categ; col++)
-                {
-                    if (!workspace.cols_possible[col + input_data.ncols_numeric]) continue;
-                    get_categs(workspace.ix_arr.data(), input_data.categ_data + input_data.nrows * col,
-                               workspace.st, workspace.end, input_data.ncat[col],
-                               model_params.missing_action, workspace.categs.data(), workspace.npresent, workspace.unsplittable);
-                    workspace.cols_possible[col + input_data.ncols_numeric] = !workspace.unsplittable;
-                    workspace.ncols_tried += !workspace.unsplittable;
-                }
-            
-
-                /* if no further splits are possible, end the procedure here */
-                workspace.npresent = workspace.ncols_tried;
-                if (!workspace.npresent) goto terminal_statistics;
-
-                /* otherwise, pick a column at random from the possible ones */
-                if (!workspace.col_sampler.max())
-                {
-                    /* no weights by columns */
-                    trees.back().col_num = std::uniform_int_distribution<size_t>
-                                                    (0, workspace.npresent - 1)
-                                                    (workspace.rnd_generator);
-                    workspace.ncols_tried = 0;
-                    for (size_t col = 0; col < input_data.ncols_tot; col++)
+                    else
                     {
+                        get_categs(workspace.ix_arr.data(), input_data.categ_data + input_data.nrows * (col - input_data.ncols_numeric),
+                                   workspace.st, workspace.end, input_data.ncat[col - input_data.ncols_numeric],
+                                   model_params.missing_action, workspace.categs.data(), workspace.npresent, workspace.unsplittable);
+                    }
 
-                        if (workspace.cols_possible[col])
+                    if (workspace.unsplittable)
+                    {
+                        workspace.cols_possible[col] = !workspace.unsplittable;
+                        workspace.col_sampler.drop_col(col);
+                    }
+
+                    else
+                    {
+                        if (col < input_data.ncols_numeric)
                         {
-                            if (workspace.ncols_tried == trees.back().col_num)
-                            {
-                                if (col < input_data.ncols_numeric)
-                                {
-                                    trees.back().col_num  = col;
-                                    trees.back().col_type = Numeric;
-                                } 
-
-                                else
-                                {
-                                    trees.back().col_num  = col - input_data.ncols_numeric;
-                                    trees.back().col_type = Categorical;
-                                }
-                                break;
-                            }
-                            workspace.ncols_tried++;
+                            trees.back().col_num  = col;
+                            trees.back().col_type = Numeric;
                         }
 
+                        else
+                        {
+                            trees.back().col_num  = col - input_data.ncols_numeric;
+                            trees.back().col_type = Categorical;
+                        }
+
+                        goto picked_col;
                     }
                 }
-
-                else
-                {
-                    /* weights by columns */
-                    std::vector<double> col_weights = workspace.col_sampler.probabilities();
-                    update_col_sampler(workspace, input_data);
-
-                    decide_column(input_data.ncols_numeric, input_data.ncols_categ,
-                                  trees.back().col_num, trees.back().col_type,
-                                  workspace.rnd_generator, workspace.runif,
-                                  workspace.col_sampler);
-                }
-
+                goto terminal_statistics;
             }
 
             /* finally, check the range if needed, and later decide on the split point */
+            picked_col:
             if (workspace.criterion == NoCrit)
                 get_split_range(workspace, input_data, model_params, trees.back());
-
         }
 
     }
@@ -577,18 +554,13 @@ void split_itree_recursive(std::vector<IsoTree>     &trees,
             add_separation_step(workspace, input_data, (double)(-1));
         
         size_t tree_from = trees.size() - 1;
-        size_t ix2, ix3;
-        std::unique_ptr<std::vector<bool>> cols_possible_ptr;
-        std::unique_ptr<std::discrete_distribution<size_t>> col_sampler_ptr;
+        std::unique_ptr<RecursionState>
+        recursion_state(new RecursionState(workspace, model_params.missing_action != Fail));
         trees.back().score = -1;
 
         /* compute statistics for NAs and remember recursion indices/weights */
-        std::unique_ptr<RecursionState> recursion_state;
         if (model_params.missing_action != Fail)
         {
-            recursion_state = std::unique_ptr<RecursionState>(new RecursionState);
-            backup_recursion_state(workspace, *recursion_state);
-
             trees.back().pct_tree_left = (long double)(workspace.st_NA - workspace.st)
                                             /
                                          (long double)(workspace.end - workspace.st + 1 - (workspace.end_NA - workspace.st_NA));
@@ -624,16 +596,6 @@ void split_itree_recursive(std::vector<IsoTree>     &trees,
             trees.back().pct_tree_left = (long double) (workspace.split_ix - workspace.st)
                                             /
                                          (long double) (workspace.end - workspace.st + 1);
-
-            ix2 = workspace.split_ix;
-            ix3 = workspace.end;
-            cols_possible_ptr  = std::unique_ptr<std::vector<bool>>(new std::vector<bool>);
-            *cols_possible_ptr = workspace.cols_possible;
-            if (workspace.col_sampler.max())
-            {
-                col_sampler_ptr  = std::unique_ptr<std::discrete_distribution<size_t>>(new std::discrete_distribution<size_t>);
-                *col_sampler_ptr = workspace.col_sampler;
-            }
             workspace.end = workspace.split_ix - 1;
         }
 
@@ -664,10 +626,9 @@ void split_itree_recursive(std::vector<IsoTree>     &trees,
 
 
         /* right branch */
+        recursion_state->restore_state(workspace);
         if (model_params.missing_action != Fail)
         {
-            restore_recursion_state(workspace, *recursion_state);
-
             switch(model_params.missing_action)
             {
                 case Impute:
@@ -695,11 +656,7 @@ void split_itree_recursive(std::vector<IsoTree>     &trees,
 
         else
         {
-            workspace.st  = ix2;
-            workspace.end = ix3;
-            workspace.cols_possible = std::move(*cols_possible_ptr);
-            if (col_sampler_ptr)
-                workspace.col_sampler = std::move(*col_sampler_ptr);
+            workspace.st = workspace.split_ix;
         }
 
         trees[tree_from].tree_right = trees.size();
