@@ -48,7 +48,6 @@
 #include <math.h>
 #include <limits.h>
 #include <string.h>
-#include <signal.h>
 #include <vector>
 #include <iterator>
 #include <numeric>
@@ -79,13 +78,11 @@
     #include <string>
     #include <fstream>
 #endif
-typedef void (*sig_t_)(int);
-#if !defined(_WIN32) && !defined(_WIN64) && !defined(_MSC_VER)
-    #include <unistd.h>
-#endif
 #ifdef _FOR_R
     #include <Rcpp.h>
 #endif
+#include <signal.h>
+typedef void (*sig_t_)(int);
 
 
 /* By default, will use Mersenne-Twister for RNG, but can be switched to something faster */
@@ -135,7 +132,7 @@ typedef void (*sig_t_)(int);
 #endif
 
 
-/*    Apple at some point decided to drop OMP library and headersfrom its compiler distribution
+/*   Apple at some point decided to drop OMP library and headers from its compiler distribution
 *    and to alias 'gcc' to 'clang', which work differently when given flags they cannot interpret,
 *    causing installation issues with pretty much all scientific software due to OMP headers that
 *    would normally do nothing. This piece of code is to allow compilation without OMP header. */
@@ -463,6 +460,22 @@ typedef struct ImputedData {
 
 } ImputedData;
 
+/* Helper when passing column weights */
+class WeightedColSampler
+{
+public:
+    std::vector<double> tree_weights;
+    size_t n;
+    size_t tree_levels;
+    size_t offset;
+    void initialize(double weights[], size_t n);
+    void drop_col(size_t col);
+    size_t sample_col(RNG_engine &rnd_generator);
+    void shuffle_cols(std::vector<size_t> &out, RNG_engine &rnd_generator);
+    bool is_initialized();
+    WeightedColSampler() = default;
+};
+
 typedef struct {
     std::vector<size_t>  ix_arr;
     std::vector<size_t>  ix_all;
@@ -474,6 +487,7 @@ typedef struct {
     size_t               st_NA;
     size_t               end_NA;
     size_t               split_ix;
+    bool                 go_to_shuffle;
     std::unordered_map<size_t, double> weights_map;
     std::vector<double>  weights_arr;    /* when not ignoring NAs and when using weights as density */
     double               xmin;
@@ -486,7 +500,7 @@ typedef struct {
     int                  ncat_tried;
     std::vector<bool>    cols_possible;
     std::vector<double>  btree_weights;  /* only when using weights for sampling */
-    std::discrete_distribution<size_t> col_sampler; /* columns can get eliminated, keep a copy for each thread */
+    WeightedColSampler   col_sampler; /* columns can get eliminated, keep a copy for each thread */
 
     /* for split criterion */
     std::vector<double>  buffer_dbl;
@@ -546,17 +560,24 @@ typedef struct WorkerForSimilarity {
     bool                assume_full_distr; /* doesn't need to have one copy per worker */
 } WorkerForSimilarity;
 
-typedef struct {
+class RecursionState {
+public:
     size_t  st;
     size_t  st_NA;
     size_t  end_NA;
     size_t  split_ix;
     size_t  end;
+    bool    go_to_shuffle;
+    bool    full_state;
     std::vector<size_t> ix_arr;
     std::vector<bool>   cols_possible;
+    WeightedColSampler  col_sampler;
     std::unique_ptr<double[]> weights_arr;
-    std::discrete_distribution<size_t> col_sampler;
-} RecursionState;
+
+    RecursionState() = default;
+    RecursionState(WorkerMemory &workspace, bool full_state);
+    void restore_state(WorkerMemory &workspace);
+};
 
 /* Function prototypes */
 
@@ -752,14 +773,13 @@ size_t check_for_missing(PredictionData  &prediction_data,
 /* helpers_iforest.cpp */
 void decide_column(size_t ncols_numeric, size_t ncols_categ, size_t &col_chosen, ColType &col_type,
                    RNG_engine &rnd_generator, std::uniform_int_distribution<size_t> &runif,
-                   std::discrete_distribution<size_t> &col_sampler);
+                   WeightedColSampler &col_sampler);
 void add_unsplittable_col(WorkerMemory &workspace, IsoTree &tree, InputData &input_data);
 void add_unsplittable_col(WorkerMemory &workspace, InputData &input_data);
-bool check_is_not_unsplittable_col(WorkerMemory &workspace, IsoTree &tree, InputData &input_data);
+bool check_is_splittable_col(WorkerMemory &workspace, IsoTree &tree, InputData &input_data);
 void get_split_range(WorkerMemory &workspace, InputData &input_data, ModelParams &model_params, IsoTree &tree);
 void get_split_range(WorkerMemory &workspace, InputData &input_data, ModelParams &model_params);
 int choose_cat_from_present(WorkerMemory &workspace, InputData &input_data, size_t col_num);
-void update_col_sampler(WorkerMemory &workspace, InputData &input_data);
 bool is_col_taken(std::vector<bool> &col_is_taken, std::unordered_set<size_t> &col_is_taken_s,
                   InputData &input_data, size_t col_num, ColType col_type);
 void set_col_as_taken(std::vector<bool> &col_is_taken, std::unordered_set<size_t> &col_is_taken_s,
@@ -768,8 +788,6 @@ void add_separation_step(WorkerMemory &workspace, InputData &input_data, double
 void add_remainder_separation_steps(WorkerMemory &workspace, InputData &input_data, long double sum_weight);
 void remap_terminal_trees(IsoForest *model_outputs, ExtIsoForest *model_outputs_ext,
                           PredictionData &prediction_data, sparse_ix *restrict tree_num, int nthreads);
-void backup_recursion_state(WorkerMemory &workspace, RecursionState &recursion_state);
-void restore_recursion_state(WorkerMemory &workspace, RecursionState &recursion_state);
 
 
 /* utils.cpp */
@@ -852,6 +870,8 @@ size_t center_NAs(size_t *restrict ix_arr, size_t st_left, size_t st, size_t cur
 void todense(size_t ix_arr[], size_t st, size_t end,
              size_t col_num, double *restrict Xc, sparse_ix Xc_ind[], sparse_ix Xc_indptr[],
              double *restrict buffer_arr);
+bool check_indices_are_sorted(sparse_ix indices[], size_t n);
+void sort_csc_indices(double *restrict Xc, sparse_ix *restrict Xc_ind, sparse_ix *restrict Xc_indptr, size_t ncols_numeric);
 
 /* mult.cpp */
 void calc_mean_and_sd(size_t ix_arr[], size_t st, size_t end, double *restrict x,
@@ -913,7 +933,7 @@ void merge_models(IsoForest*     model,      IsoForest*     other,
                   ExtIsoForest*  ext_model,  ExtIsoForest*  ext_other,
                   Imputer*       imputer,    Imputer*       iother);
 
-#ifdef _ENABLE_CEREAL
+#if defined(_ENABLE_CEREAL) || defined(_FOR_PYTHON)
 /* serialize.cpp */
 void serialize_isoforest(IsoForest &model, std::ostream &output);
 void serialize_isoforest(IsoForest &model, const char *output_file_path);
@@ -950,8 +970,9 @@ void deserialize_ext_isoforest(ExtIsoForest &output_obj, void *input_file_path);
 void serialize_imputer(Imputer &imputer, void *output_file_path);
 void deserialize_imputer(Imputer &output_obj, void *input_file_path);
 bool py_should_use_char();
+bool has_cereal();
 #endif /* _FOR_PYTHON */
-#endif /* _ENABLE_CEREAL */
+#endif /* _ENABLE_CEREAL || _FOR_PYTON */
 
 /* sql.cpp */
 std::vector<std::string> generate_sql(IsoForest *model_outputs, ExtIsoForest *model_outputs_ext,
