@@ -635,6 +635,114 @@ void weighted_shuffle(size_t *restrict outp, size_t n, double *restrict weights,
 
 }
 
+/* This one samples with replacement. Algorithm is the same as above but keeps
+   the weights after taking each sample. */
+void WeightedColSampler::initialize(double weights[], size_t n)
+{
+    this->n = n;
+    this->tree_levels = log2ceil(n);
+    if (!this->tree_weights.size())
+        this->tree_weights.resize(pow2(this->tree_levels + 1), 0);
+    else {
+        if (this->tree_weights.size() != pow2(this->tree_levels + 1))
+            this->tree_weights.resize(this->tree_levels);
+        std::fill(this->tree_weights.begin(), this->tree_weights.end(), 0.);
+    }
+
+    /* compute sums for the tree leaves at each node */
+    this->offset = pow2(this->tree_levels) - 1;
+    for (size_t ix = 0; ix < this->n; ix++) {
+        this->tree_weights[ix + this->offset] = weights[ix];
+    }
+    for (size_t ix = pow2(this->tree_levels+1) - 1; ix > 0; ix--) {
+        this->tree_weights[ix_parent(ix)] += tree_weights[ix];
+    }
+}
+
+void WeightedColSampler::drop_col(size_t col)
+{
+    if (this->is_initialized())
+    {
+        size_t curr_ix = col + this->offset;
+        if (this->tree_weights[curr_ix] > 0)
+        {
+            this->tree_weights[curr_ix] = 0.;
+            for (size_t lev = 0; lev < this->tree_levels; lev++)
+            {
+                curr_ix = ix_parent(curr_ix);
+                this->tree_weights[curr_ix] =   this->tree_weights[ix_child(curr_ix)]
+                                              + this->tree_weights[ix_child(curr_ix) + 1];
+            }
+        }
+    }
+}
+
+size_t WeightedColSampler::sample_col(RNG_engine &rnd_generator)
+{
+    size_t curr_ix = 0;
+    double rnd_subrange, w_left;
+    double curr_subrange = this->tree_weights[0];
+    for (size_t lev = 0; lev < tree_levels; lev++)
+    {
+        rnd_subrange = std::uniform_real_distribution<double>(0., curr_subrange)(rnd_generator);
+        w_left = this->tree_weights[ix_child(curr_ix)];
+        curr_ix = ix_child(curr_ix) + (rnd_subrange >= w_left);
+        curr_subrange = this->tree_weights[curr_ix];
+    }
+
+    return curr_ix - this->offset;
+}
+
+void WeightedColSampler::shuffle_cols(std::vector<size_t> &out, RNG_engine &rnd_generator)
+{
+    out.reserve(this->n);
+    out.resize(this->n);
+    std::vector<double> curr_weight = this->tree_weights;
+    size_t n_available = 0;
+    double rnd_subrange, w_left;
+    double curr_subrange;
+    int curr_ix;
+
+    for (size_t el = 0; el < this->n; el++)
+    {
+        /* go down the tree by drawing a random number and
+           checking if it falls in the left or right sub-ranges */
+        curr_ix = 0;
+        curr_subrange = curr_weight[0];
+        if (curr_subrange == 0)
+            break;
+
+        for (size_t lev = 0; lev < this->tree_levels; lev++)
+        {
+            rnd_subrange = std::uniform_real_distribution<double>(0., curr_subrange)(rnd_generator);
+            w_left = curr_weight[ix_child(curr_ix)];
+            curr_ix = ix_child(curr_ix) + (rnd_subrange >= w_left);
+            curr_subrange = curr_weight[curr_ix];
+        }
+
+        /* finally, add element from this iteration */
+        out[el] = curr_ix - this->offset;
+
+        /* now remove the weight of the chosen element */
+        curr_weight[curr_ix] = 0;
+        for (size_t lev = 0; lev < this->tree_levels; lev++)
+        {
+            curr_ix = ix_parent(curr_ix);
+            curr_weight[curr_ix] =   curr_weight[ix_child(curr_ix)]
+                                   + curr_weight[ix_child(curr_ix) + 1];
+        }
+
+        n_available++;
+    }
+
+    out.resize(n_available);
+}
+
+bool WeightedColSampler::is_initialized()
+{
+    return this->tree_weights.size() > 0;
+}
+
 /* For hyperplane intersections */
 size_t divide_subset_split(size_t ix_arr[], double x[], size_t st, size_t end, double split_point)
 {
@@ -768,6 +876,7 @@ void divide_subset_split(size_t ix_arr[], size_t st, size_t end, size_t col_num,
                         st++;
                     }
                     if (curr_pos == end_col && row < ix_arr + end)
+                    {
                         for (size_t *r = row + 1; r <= ix_arr + end; r++)
                         {
                             temp       = ix_arr[st];
@@ -775,6 +884,7 @@ void divide_subset_split(size_t ix_arr[], size_t st, size_t end, size_t col_num,
                             *r         = temp;
                             st++;
                         }
+                    }
                     if (row == ix_arr + end || curr_pos == end_col) break;
                     curr_pos = std::lower_bound(Xc_ind + curr_pos + 1, Xc_ind + end_col + 1, *(++row)) - Xc_ind;
                 }
@@ -1277,8 +1387,8 @@ void get_range(size_t ix_arr[], double x[], size_t st, size_t end,
     {
         for (size_t row = st; row <= end; row++)
         {
-            xmin = fmin(xmin, x[ix_arr[row]]);
-            xmax = fmax(xmax, x[ix_arr[row]]);
+            xmin = std::fmin(xmin, x[ix_arr[row]]);
+            xmax = std::fmax(xmax, x[ix_arr[row]]);
         }
     }
 
@@ -1355,8 +1465,8 @@ void get_range(size_t ix_arr[], size_t st, size_t end, size_t col_num,
             if (Xc_ind[curr_pos] == *row)
             {
                 nmatches++;
-                xmin = fmin(xmin, Xc[curr_pos]);
-                xmax = fmax(xmax, Xc[curr_pos]);
+                xmin = std::fmin(xmin, Xc[curr_pos]);
+                xmax = std::fmax(xmax, Xc[curr_pos]);
                 if (row == ix_arr + end || curr_pos == end_col) break;
                 curr_pos = std::lower_bound(Xc_ind + curr_pos, Xc_ind + end_col + 1, *(++row)) - Xc_ind;
             }
@@ -1374,8 +1484,8 @@ void get_range(size_t ix_arr[], size_t st, size_t end, size_t col_num,
 
     if (nmatches < (end - st + 1))
     {
-        xmin = fmin(xmin, 0);
-        xmax = fmax(xmax, 0);
+        xmin = std::fmin(xmin, 0);
+        xmax = std::fmax(xmax, 0);
     }
     unsplittable = (xmin == xmax) || (xmin == HUGE_VAL && xmax == -HUGE_VAL);
 
@@ -1543,6 +1653,52 @@ void todense(size_t ix_arr[], size_t st, size_t end,
     }
 }
 
+/* TODO: remove this once SciPy implements it for 'size_t' */
+bool check_indices_are_sorted(sparse_ix indices[], size_t n)
+{
+    if (n <= 1)
+        return true;
+    if (indices[n-1] < indices[0])
+        return false;
+    for (size_t ix = 1; ix < n; ix++)
+        if (indices[ix] < indices[ix-1])
+            return false;
+    return true;
+}
+
+void sort_csc_indices(double *restrict Xc, sparse_ix *restrict Xc_ind, sparse_ix *restrict Xc_indptr, size_t ncols_numeric)
+{
+    std::vector<double> buffer_sorted_vals;
+    std::vector<sparse_ix> buffer_sorted_ix;
+    std::vector<size_t> argsorted;
+    size_t n_this;
+    size_t ix1, ix2;
+    for (size_t col = 0; col < ncols_numeric; col++)
+    {
+        ix1 = Xc_indptr[col];
+        ix2 = Xc_indptr[col+1];
+        n_this = ix2 - ix1;
+        if (n_this && !check_indices_are_sorted(Xc_ind + ix1, n_this))
+        {
+            if (buffer_sorted_vals.size() < n_this)
+            {
+                buffer_sorted_vals.resize(n_this);
+                buffer_sorted_ix.resize(n_this);
+                argsorted.resize(n_this);
+            }
+            std::iota(argsorted.begin(), argsorted.begin() + n_this, ix1);
+            std::sort(argsorted.begin(), argsorted.begin() + n_this,
+                      [&Xc_ind](const size_t a, const size_t b){return Xc_ind[a] < Xc_ind[b];});
+            for (size_t ix = 0; ix < n_this; ix++)
+                buffer_sorted_ix[ix] = Xc_ind[argsorted[ix]];
+            std::copy(buffer_sorted_ix.begin(), buffer_sorted_ix.begin() + n_this, Xc_ind + ix1);
+            for (size_t ix = 0; ix < n_this; ix++)
+                buffer_sorted_vals[ix] = Xc[argsorted[ix]];
+            std::copy(buffer_sorted_vals.begin(), buffer_sorted_vals.begin() + n_this, Xc + ix1);
+        }
+    }
+}
+
 /* Function to handle interrupt signals */
 void set_interrup_global_variable(int s)
 {
